from __future__ import absolute_import

import os

import afinipy.utils as uf
import afinipy.path_funcs as pf

from afinipy.module import Module


class Directory(object):
    """Class that orchastrates the parsing of a directory in
    the tree of the target directory
    """
    def __init__(self, path, mode, package, parents, files, exclude=None, cf_exclude=None, udef_exclude=None,
                 verbose=False, dry_run=False):
        """Initialise the class

        Parameters
        ----------
        path : str
            The path to the root of the directory
        mode : str
            top_level or recursive
        package : str
            the prefix for the import statement ['.', 'package_name']
        parents : str
            The parent directories starting at the target path
        files : list
            The files in the directory
        exclude : list-like
            The modules to exclude
        cf_exclude : str, None
            Exclude `classes` or `functions`
        udef_exclude : set
            Set of functions or classes to be excluded
        verbose : bool
            Print the import statements
        dry_run : bool
            do not write, only print
        """
        # get absolute path and check existence
        self.root = pf.adir(path)

        # top_level or recursive
        self.mode = mode

        # the prefix for the print statements
        self.package = package

        # name of the class is the directory name
        self.name = pf.dir_name(self.root)

        # the name(s) of the parent directories seperated by a dot
        self.parents = parents

        # files contains all files found
        # in the root directory
        self.files = files

        # the modules that should be excluded
        self.exclude = exclude or set()
        self.exclude.update({'__init__', 'setup'})

        self.cf_exclude = cf_exclude
        self.udef_exclude = udef_exclude

        # print import blocks
        self.verbose = verbose

        # Don't write only print
        self.dry_run = dry_run
        if self.dry_run:
            self.verbose = True

        # files parser sets all the python modules found
        # in the directory and creates module objects for each
        self.files_parser()

        if self.mode == 'recursive':
            self.build_init()
            if not self.dry_run:
                self.write_init()

    def _ismodule(self, f):
        """Check if file is python module

        Parameters
        ----------
        f : str
            path to file

        Returns
        -------
        bool
            whether the file is a python module that is relevant
        """
        if os.path.isfile(os.path.join(self.root, f)):
            name, ext = os.path.splitext(f)
            return ((ext == '.py') and
                    (name[0] not in {'.', '_'}) and
                    (name not in (self.exclude)))
        else:
            return False

    def files_parser(self):
        """Filter the python modules from the files and set as modules,
        where we exclude dotfiles, files starting with `_` and
        __init__ files.

        For each module we create a Module instance that extracts the
        functions and classes without importing and correcting for
        star imports in the module
        """
        # create the list of module names
        self.module_names = set([os.path.splitext(f)[0] for f in self.files if self._ismodule(f)])

        self.modules = [
            Module(
                name=module,
                path=os.path.join(self.root, module) + '.py',
                parents=self.parents,
                exclude=self.cf_exclude,
                udef_exclude=self.udef_exclude)
            for module in self.module_names
        ]

    def build_init(self):
        """Collect udefs from modules and create all_udefs and imports

        We create a string that contains the import statements sorted
        alphabetically per module and within modules. A module's
        statements are seperated by a newline
        """
        # if recursive and no package is given we strip the parents
        if self.package == '':
            self.parents = ''

        self.imports = []
        self.all_udefs = []
        for module in sorted(self.modules, key=lambda x: x.name.lower()):
            block = []

            # collect all the udefs for __all__
            self.all_udefs.extend(module.udefs)

            for udef in sorted(module.udefs, key=lambda x: x.lower()):
                # Template imports
                block.append('from {0} import {1}\n'.format(uf.concat_imports((self.parents, module.name)), udef))
            # create block of import statements per module seperated by empty line
            self.imports.append(''.join(block))
        self.imports = '\n'.join(self.imports)
        if self.verbose:
            print('Import statements directory: ', self.name)
            print(self.imports)
            print('\n__all__ = {}\n'.format(sorted(self.all_udefs, key=lambda x: x.lower())))

    def write_init(self):
        """Write the imports string and all_udefs list to the init file
        """
        with open(os.path.join(self.root, '__init__.py'), 'w+') as init_file:
            # Write the __init__, create if it doesn't exist
            init_file.write(self.imports)

            # The objects that will be added for __all__
            # Template for __all__ functions in all modules
            init_file.write('\n__all__ = {}\n'.format(sorted(self.all_udefs, key=lambda x: x.lower())))
        init_file.close()
        print('{} has been appended/generated\n'.format(os.path.join(self.root, '__init__.py')))



def concat_imports(values, sep='.'):
    """Combine import statements

    Parameters
    ----------
    values : list-like
        The values to be concatenated
    sep : string
        The value between the statements

    Returns
    -------
    str
        The concatenated string
    """
    if len(values) >= 1:
        store = []
        for v in values:
            if isinstance(v, (list, tuple)):
                store.extend(v)
            else:
                store.append(v)
        return sep.join(store)
    else:
        return ''


def notin(validate, baseline):
    """Determine wich elements are not present in the baseline
    but are present in the `validate set`

    Parameters
    ----------
    validate : set, or list-like
        The set from which we want to find the elements
    baseline : set, or list-like
        The set that serves as a baseline set

    Returns
    -------
    set
        Set containing the element not in baseline
    """
    return set(validate) - set(baseline)


def ordered_notin(validate, baseline):
    """Determine wich elements are not present in the baseline
    but are present in the `validate set`

    Parameters
    ----------
    validate : list
        The list from which we want to find the elements
    baseline : list
        The list that serves as a baseline set

    Returns
    -------
    set
        Set containing the element not in baseline
    """
    return [v for v in validate if v not in baseline]



from __future__ import absolute_import

import click

from afinipy.base import Afinipy


@click.command()
@click.argument('path', type=click.Path(exists=True), required=True)
@click.option('--mode', '-m', default='top_level', help='Type of __init__(s) to build. Default is "top_level"')
@click.option('--package', '-P', default='', help='Name of the package the init will be build for')
@click.option('--exclude', '-e', default=None, help='Exclude all functions or classes')
@click.option('--exclusion_path', '-E', default=None, type=click.Path(exists=True),
              required=False, help='Path to exclusion file')
@click.option('--verbose', '-v', is_flag=True, default=False, help='Print import statements')
@click.option('--dry_run', is_flag=True, default=False, help='Do not write, only print')
def cli(path, mode, package, exclude, exclusion_path, verbose, dry_run):
    """Build __init__ file for the package

    Parameters
    ----------
    path : str
        The path to the root of the directory
    mode : str
        The mode to run afinipy in. Options are:
        - top_level: everything is set in top level __init__
        - recursive: each directory level gets own __init__
        Optional, default is `top_level`
    package : str
        Package name; default is `None`
    exclude : str, optional
        Exclude all functions or classes
    exclusion_path : str, optional
        Path to the file containing the exclusions
    verbose : bool
        Print the import statements
    dry_run : bool
        Do not write the __init__ but only print
    """
    # Obtain path to target directory
    afinipy = Afinipy(
        path,
        mode=mode,
        package=package,
        exclude=exclude,
        exclusion_path=exclusion_path,
        verbose=verbose,
        dry_run=dry_run
    )
    afinipy.build_init()


if __name__ == '__main__':
    cli()



from __future__ import absolute_import

import os

from afinipy.exceptions import DirectoryNotFound


def path_list(path):
    """The path to the directory split on os.path.sep

    Parameters
    ----------
    path : str
        Path to directory

    Returns:
    list
        list of names of the directories in the path
    """
    return path.split(os.path.sep)


def dir_depth(path, base_depth=None):
    """The number of directories below the system root. If base_depth
    is passed the relative directory level is returned

    Parameters
    ----------
    path : str
        Path to directory
    base_depth : int/float
        The dir_depth of the relative root

    Returns:
    int
        the absolute or relative dir level
    """
    if isinstance(base_depth, (int, float)):
        return len(path.split(os.path.sep)) - base_depth
    else:
        return len(path.split(os.path.sep))


def dir_name(path):
    """Get name of the directory

    Parameters
    ----------
    path : str
        Path to directory

    Returns
    -------
    str
        The name of the directory
    """
    return os.path.split(path)[-1]


def adir(path):
    """Get full path to dir

    Parameters
    ----------
    path : str
        Path to directory

    Returns
    -------
    str
        The absolute path to dir

    Raises
    ------
    DirectoryNotFound
        If the path does not exist
    """
    apath = os.path.abspath(path)
    if not os.path.isdir(apath):
        raise DirectoryNotFound(path)
    return apath


def wdir():
    """Get absolute path to working directory

    Returns
    -------
    str
        abs path to working directory
    """
    return os.getcwd()


def mdir():
    """Get absolute path to module/notebook directory

    Returns
    -------
    str
        abs path to dir where file is located
    """
    try:
        p = os.path.realpath(os.path.dirname(__main__))
    except NameError:
        p = os.path.realpath(os.path.dirname('__main__'))
    return p


def pdir():
    """Get absolute path to parent directory

    Returns
    -------
    str
        abs path to parent directory
    """
    return os.path.realpath(os.path.pardir)


def sdir(dirname, filename=None):
    """Get path to sibling directory of name `dirname`

    Parameters
    ----------
    dirname : str | tuple | list
        The name of the sibling directory
    filename : str
        The name of the file in the sibling directory

    Returns
    -------
    str
        path to sibling directory or file
    """
    if isinstance(dirname, (tuple, list)):
        dirname = os.path.join(*dirname)
    if filename:
        return os.path.join(pdir(), dirname, filename)
    else:
        return os.path.join(pdir(), dirname)



from __future__ import absolute_import

import os

import afinipy.path_funcs as pf
import afinipy.utils as uf

from afinipy.directory import Directory
from afinipy.exclusion_paser import ExclusionPaser

from afinipy.exceptions import IllegalSetting
from afinipy.exceptions import WrongSettingsType


class Afinipy(object):
    """Base class that orchastrates the parsing and creation of
    the __init__ file
    """
    def __init__(self, path, **kwargs):
        """Initialise the class

        Parameters
        ----------
        path : str
            The path to the root of the directory
        mode : str
            The mode to run afinipy in. Options are:
            - top_level: everything is set in top level __init__
            - recursive: each directory level gets own __init__
            Optional, default is `top_level`
        package : str
            Package name; default is `None`
        exclude : str, optional
            Exclude all functions or classes
        exclusion_path : str, optional
            Path to the file containing the exclusions
        verbose : bool
            Print the import statements
        dry_run : bool
            Do not write the __init__ but only print
        """
        # validate type and value of parameters
        if not isinstance(path, str):
            raise WrongSettingsType('path', path, str)

        # get absolute path and check existence
        self._base_path = pf.adir(path)

        # The way to build the inits
        self._mode = kwargs.pop('mode', 'top_level')

        # validate mode setting
        if (self._mode is not None) and (self._mode not in {'top_level', 'recursive'}):
            raise IllegalSetting('mode', self._mode)

        # The prefix for the import statements
        self._package = kwargs.pop('package', '')

        self.exclude = kwargs.pop('exclude', None)

        self._verbose = kwargs.pop('verbose', False)
        self._dry_run = kwargs.pop('dry_run', False)

        # exclusion_path to the directories the user wants excluded
        self.exclusion_path = kwargs.pop('exclusion_path', None)

        if self.exclusion_path:
            ep = ExclusionPaser(self.exclusion_path)
            exclusions = ep.exclusion_parser()

            self._dir_exclude = exclusions['direc']
            self._module_exclude = exclusions['module']
            self._udef_exclude = exclusions['udef']
        else:
            self._dir_exclude = set()
            self._module_exclude = set()
            self._udef_exclude = set()

        self._name = pf.dir_name(self._base_path)

        # the number of directories we are from system root
        self._base_depth = pf.dir_depth(self._base_path)

        # list of the names of the directories from root
        self._base_path_list = pf.path_list(self._base_path)

        # initialise the directory collection where the target directory
        # is the root directory
        self.dirs = []

    def build_init(self, **kwargs):
        """Create the init file(s)

        Run the directory parser which will build inits per
        directory if mode is recursive else collect all statements
        and create the init

        Parameters
        ----------
        verbose : bool
            Print import statements; default is `False`
        dry_run : bool
            do not write, only print
        """
        self._verbose = self._verbose or kwargs.pop('verbose', False)

        self._dry_run = self._dry_run or kwargs.pop('dry_run', False)
        if self._dry_run:
            self._verbose = True

        # reset self.dirs to prevent extending
        self.dirs = []

        # start parsing
        self.directory_parser()

        if self._mode == 'top_level':
            self.top_level()
            if not self._dry_run:
                self.write_init()

    def _get_parents(self, path):
        """Get and concat parents to directory

        Parameters
        ----------
        path : str
            The path to the directory

        Returns
        -------
        str
            The parents concatonated with seperator sep
        """
        return uf.concat_imports((self._package, pf.path_list(path)[self._base_depth:]))

    def _exclude_dir(self, d):
        """Determine if the directory is relevant

        Parameters
        ----------
        d : str
            the name of the directory

        Returns
        -------
        bool
            whether to exclude the directory
        """
        return (d[0] in {'_', '.'}) or (d in self._dir_exclude)

    def directory_parser(self):
        """Recursively search the target directory and create collection
        all directory classes with the python modules in that directory
        as attribute
        """

        # Loop through all directories in and below base_path
        for root, dirs, files in os.walk(self._base_path):
            # extract the name of current directory
            dname = pf.dir_name(root)
            # check if directory should be excluded
            if self._exclude_dir(dname):
                continue
            else:
                self.dirs.append(
                    Directory(
                        path=root,
                        mode=self._mode,
                        package=self._package,
                        parents=self._get_parents(root),
                        files=files,
                        exclude=self._module_exclude,
                        cf_exclude=self.exclude,
                        udef_exclude=self._udef_exclude,
                        verbose=self._verbose,
                        dry_run=self._dry_run
                    )
                )

    def write_init(self):
        """Write the imports string and all_udefs list to the init file

        """
        with open(os.path.join(self._base_path, '__init__.py'), 'w+') as init_file:
            # Write the __init__, create if it doesn't exist
            init_file.write(self.imports)

            # The objects that will be added for __all__
            # Template for __all__ functions in all modules
            init_file.write('\n__all__ = {}\n'.format(sorted(self.all_udefs, key=lambda x: x.lower())))
        init_file.close()
        print('{} has been appended/generated\n'.format(os.path.join(self._base_path, '__init__.py')))

    def top_level(self):
        """Create __init__ at root level with all user defined functions
        and classes in the directory tree
        """
        self.imports = []
        self.all_udefs = []
        for direc in self.dirs:
            for module in sorted(direc.modules, key=lambda x: x.name.lower()):
                self.all_udefs.extend(module.udefs)
                block = []
                for udef in sorted(module.udefs, key=lambda x: x.lower()):
                    # Template imports
                    block.append('from {0} import {1}\n'.format(
                        uf.concat_imports((module.parents, module.name)),
                        udef
                        )
                    )
                self.imports.append(''.join(block))
        self.imports = '\n'.join(self.imports)
        if self._verbose:
            print('Import statements directory: ', self._name)
            print(self.imports)
            print('\n__all__ = {}\n'.format(sorted(self.all_udefs, key=lambda x: x.lower())))



from __future__ import absolute_import

import os

from afinipy.exceptions import FileNotFound


class ExclusionPaser(object):
    """Class that orchastrates the parsing of the exlusion path"""
    def __init__(self, path):
        """Initialise the class

        Parameters
        ----------
        path : str
            The absolute path to the module includign extension
        """
        # absolute path to module with extension
        self.path = os.path.realpath(path)

        if not os.path.isfile(self.path):
            raise FileNotFound(self.path)

        # Track empty lines
        self.empty_lines = []
        # Initialise line store
        self.line_store = {}
        # Initialise the module store
        self.module_store = set()
        # Initialise the module store
        self.udef_store = set()
        # Initialise the module store
        self.direc_store = set()

        self.uvars = ['udef', 'udefs', 'Udef', 'Udefs']
        self.mvars = ['module', 'modules', 'Module', 'Modules']
        self.dvars = ['directories', 'Directories', 'Dirs',
                      'dirs', 'direcs', 'Direcs', 'direc', 'Direc']

    def exclusion_parser(self):
        """Parse the exclusion file"""
        with open(self.path) as f:
            self.contents = [l.rstrip('\n') for l in f]
        self._parse_lines()
        return {
            'direc': self.direc_store,
            'module': self.module_store,
            'udef': self.udef_store
        }

    def _parse_lines(self):
        """Parse lines and create udef, module and directory store"""
        for i, line in enumerate(self.contents):
            if line == '':
                self.empty_lines.append(i)
            else:
                self.line_store[i] = line
                self.search_keyswords(self.uvars, self.udef_store)
                self.search_keyswords(self.mvars, self.module_store)
                self.search_keyswords(self.dvars, self.direc_store)

    def search_keyswords(self, options, store):
        """Search for keywoards in the lines and add to the respective store"""
        for number, line in self.line_store.items():
            for kw in options:
                defin, content = line.split(':')
                if kw == defin:
                    for k in content.split(','):
                        store.add(k.strip())



from __future__ import absolute_import

import ast

from afinipy.exceptions import IllegalSetting


class Module(object):
    """Class that orchastrates the parsing of a module in
    the tree of the target directory
    """
    def __init__(self, name, path, parents, udef_exclude, exclude=None):
        """Initialise the class

        Parameters
        ----------
        name : str
            The name of the module
        path : str
            The absolute path to the module includign extension
        parents : str
            The parent directories starting at the target path
        udef_exclude : set
            Set of classes or functions that are to be excluded
        exclude : None, str
            Exclude classes or functions.
        """

        # name of the class is the module name
        self.name = name

        # absolute path to module with extension
        self.path = path

        # the name(s) of the parent directories seperated by a dot
        self.parents = parents

        # whether to exclude functions, classes
        self.exclude = exclude

        # set of functions/classes to exclude
        self.udef_exclude = udef_exclude

        # module parser sets all the functions and classes found in
        # module and exclude those that are marked private
        self.module_parser()

    def _extract(self, v):
        """Determine is value should be extracted

        Parameters
        ----------
        v : ast.child_node

        Returns
        -------
        bool
            Whether the value should be extracted
        """
        return (isinstance(v, (ast.ClassDef, ast.FunctionDef)) and (v.name[0] != '_'))

    def _extract_classes(self, v):
        """Determine is value should be extracted

        Parameters
        ----------
        v : ast.child_node

        Returns
        -------
        bool
            Whether the value should be extracted
        """
        return (isinstance(v, ast.ClassDef) and (v.name[0] != '_'))

    def _extract_funcs(self, v):
        """Determine is value should be extracted

        Parameters
        ----------
        v : ast.child_node

        Returns
        -------
        bool
            Whether the value should be extracted
        """
        return (isinstance(v, ast.FunctionDef) and (v.name[0] != '_'))

    def module_parser(self):
        """Extract both classes and functions from module"""

        # parse the file and create a syntax tree
        with open(self.path) as fh:
            mroot = ast.parse(fh.read(), self.path)

        if self.exclude is None:
            # Extract both classes and functions from module, excluding
            # those that begin with `_`
            self.udefs = [n.name for n in ast.iter_child_nodes(mroot) if
                          (self._extract(n)) and (n.name not in self.udef_exclude)]

        elif self.exclude == 'functions':
            # Extract classes from module, excluding
            # those that begin with `_`
            self.udefs = [n.name for n in ast.iter_child_nodes(mroot) if
                          (self._extract_classes(n)) and (n.name not in self.udef_exclude)]

        elif self.exclude == 'classes':
            # Extract functions from module, excluding
            # those that begin with `_`
            self.udefs = [n.name for n in ast.iter_child_nodes(mroot) if
                          (self._extract_funcs(n)) and (n.name not in self.udef_exclude)]

        else:
            raise IllegalSetting('exclude', self.exclude)



from setuptools import setup, find_packages

setup(
    name='afinipy',
    version='0.1.3',
    description='Automated init builder',
    author='Ralph Urlus',
    author_email='rurlus.dev@gmail.com',
    packages=find_packages(exclude=('tests',)),
    install_requires=[
        'Click',
    ],
    entry_points='''
        [console_scripts]
        afinipy=afinipy.afinipy:cli
    ''',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/RUrlus/afinipy/issues',
        'Source': 'https://github.com/RUrlus/afinipy',
    },
    keywords='init development',
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



class IllegalSetting(Exception):
    """Exception raised when unrecognised setting is passed"""
    def __init__(self, setting, value):
        """Raise IllegalSetting exception

        Parameters
        ----------
        setting : str
            The name of the setting
        value : T
            The value passed for the setting
        """
        m = "The vale for setting {0} is not unrecognised, value: {1}".format(setting, value)
        Exception.__init__(self, m)


class WrongSettingsType(Exception):
    """Exception raised if setting is of wrong type"""
    def __init__(self, setting, value, correct_type=None):
        """Raise the exception

        Parameters
        ----------
        setting : str
            The name of the setting
        value : T
            The value passed for the setting
        correct_type : T
            The expected type of the value, optional
        """
        if correct_type:
            m = 'Setting {0} is not of type {1} but {2}'.format(setting, correct_type, type(value))
        else:
            m = 'Setting {0} is of wrong type {1}'.format(setting, type(value))
        Exception.__init__(self, m)


class DirectoryNotFound(Exception):
    """Exception raised if directory cannot be found"""
    def __init__(self, path):
        """Raise the exception

        Parameters
        ----------
        path : str
            path to target directory
        """
        m = 'Directory {} does not exist, please specify the full path'
        Exception.__init__(self, m)


class FileNotFound(Exception):
    """Exception raised if file cannot be found"""
    def __init__(self, path):
        """Raise the exception

        Parameters
        ----------
        path : str
            path to target file
        """
        m = 'File {} does not exist, please specify the full path'
        Exception.__init__(self, m)
